#include <vdr/skins.h>
#include <vdr/videodir.h>
#include "dvbchain.h"
#include "tstools.h"

// --- cRecorderChain --------------------------------------------------------

cRecorderChain* cRecorderChain::Create(cDevice* Device, cTimer* Timer)
{
    cRecording recording(Timer, Timer->Event());
    char const* fileName = recording.FileName();
    isyslog("recording %s", fileName);

    if (!MakeDirs(fileName, true))
        return NULL;

    cRecorderChain* chain = new cRecorderChain(recording.FileName(), Timer->Channel(), Timer->Priority());
    if (!Device->AttachReceiver(chain)) {
        delete chain;
        return NULL;
    }

    return chain;
}

cRecorderChain::cRecorderChain(char const* FileName, cChannel const* Channel, int Priority)
    : cReceiver(Channel->GetChannelID(), Priority, Channel->Vpid(), Channel->Apids(), Channel->Dpids(), Channel->Spids())
    , cThread("RecorderChain")
    , fileName(FileName)
    , buffer(MEGABYTE(2), TS_SIZE * 2, true, "RecorderChain")
    , vEsExtractor(Channel->Vpid())
    , vEsIndexer(*fileName)
    , fileWriter(*fileName)
{
    buffer.SetTimeouts(0, 100);

    tsSource.Connect(&tsScanner);
    tsScanner.Connect(&vEsExtractor).Connect(&vEsIndexer);
    tsScanner.Connect(&fileWriter);
}

void cRecorderChain::Receive(uchar* Data, int Length)
{
    if (!Running())
        return;

    int Put = buffer.Put(Data, Length);
    if (Put < Length)
        buffer.ReportOverflow(Length - Put);
}

void cRecorderChain::Action()
{
	while (Running()) {
        // XXX check for buffer fill level

		int Length;
		uchar* b = buffer.Get(Length);
		if (!b) // no packets available
            continue;

        int Used = tsSource.PutPackets(b, Length);
        buffer.Del(Used);
	}
}

// --- nsRecorderChains ------------------------------------------------------

namespace nsRecorderChains
{
    cList<cRecorderChain> RecordControls;

    bool Start(cTimer* Timer, bool Pause)
    {
        if (Timer == NULL)
            Timer = new cTimer(true, Pause);

        cChannel const* channel = Timer->Channel();
        int priority = Timer->Priority();
        cDevice* device = cDevice::GetDevice(channel, priority, false);
        if (device == NULL) {
            if (Timer == NULL || (priority >= Setup.PrimaryLimit && !Timer->Pending())) {
                Skins.Message(mtError, "No free DVB device to record!");
                isyslog("no free DVB device to record channel %d", channel->Number());
            }
            return false;
        }

        dsyslog("switching device %d to channel %d", device->DeviceNumber() + 1, channel->Number());
        if (!device->SwitchChannel(channel, false)) {
            Skins.Message(mtError, "Couldn't switch device!");
            esyslog("ERROR: couldn't switch device %d to channel %d", device->DeviceNumber() + 1, channel->Number());
            return false;
        }

        if (!Timer->Matches()) { // XXX
            dsyslog("Timer not matching (?)");
            return false;
        }

        cRecorderChain* chain = cRecorderChain::Create(device, Timer);
        if (chain == NULL)
            return false;

        RecordControls.Add(chain);
        return true;
    }
}

// --- cPlayerChain ----------------------------------------------------------

static int apids[] = { 256, 0 };

cPlayerChain::cPlayerChain(char const* FileName)
    : cThread("PlayerChain")
    , cControl(this)
    , fileName(FileName, false)
    , index(FileName, false)
    , recordFile(fileName.Open())
    , pesRemuxer(255, apids, NULL)
{
    tsSource.Connect(&pesRemuxer).Connect(this);
        //.Connect(&tsScanner)
}

void cPlayerChain::ReceivePackets(uchar const* Data, int Length)
{
    cPoller poller;
    if (DevicePoll(poller, 100)) {
        PlayPes(Data, Length);
    }
}

void cPlayerChain::Action()
{
    uchar buffer[TS_SIZE * 100];
    ssize_t bufferUsed = 0;

    while (Running()) {
        ssize_t r;
        if (recordFile != NULL) {
            if ((r = recordFile->Read(buffer + bufferUsed, sizeof(buffer) - bufferUsed)) < 0) {
                esyslog("ERROR: couldn't read from recording");//XXX
                break;
            } else if (r == 0)
                recordFile = NULL;
            bufferUsed += r;
        }

        int Used = tsSource.PutPackets(buffer, bufferUsed);
        bufferUsed -= Used;
        if (bufferUsed > 0)
            memmove(buffer, buffer + Used, bufferUsed);

        if (r == 0 && bufferUsed < TS_SIZE)
            break;
    }
}
