#ifndef __DVBSOURCE_H
#define __DVBSOURCE_H

#include <vdr/channels.h>
#include <vdr/receiver.h>
#include <vdr/ringbuffer.h>
#include <vdr/thread.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include "dvbpacket.h"
#include "dvbsink.h"
#include "streamtypes.h"
#include "tstools.h"

template<class Input, class Output>
class cDvbFilter;

inline int GetCaValue(tChannelID const& channelID)
{
    if (channelID == tChannelID::InvalidID)
        return 0;
    cChannel* Channel = Channels.GetByChannelID(channelID);
    if (Channel == 0)
        return 0;
    return Channel->Ca();
}

// --- cDvbSource ------------------------------------------------------------

template<class Output>
class cDvbSource
{
public:
    void Connect(cDvbSink<Output>* Sink) { sinks.Add(Sink); Sink->SetDvbPacketInfos(GetPacketInfos()); }
        ///< Connects an appropriate sink to this source. When EmitPackets()
        ///< is called by an implementation, all connected sinks will be
        ///< notified. Notice that connected sinks WILL NOT BE destroyed when
        ///< the source gets destroyed!
    
    template<class Input>
    cDvbSource<Input>& Connect(cDvbFilter<Output, Input>* Filter) { Connect(static_cast<cDvbSink<Output>*>(Filter)); return *Filter; }
        ///< This form of Connect enables chaining filters (that are both 
        ///< source and sink) like this:
        ///< source.Connect(filter1).Connect(filter2).Connect(sink);

protected:
    cDvbSource() {}

    virtual ~cDvbSource() = 0;

    void EmitPackets(uchar const* Data, int Length);
        ///< Must be called by any implementation when packets of the requested
        ///< flavour are available.

private:
    virtual tDvbPacketInfos& GetPacketInfos() = 0;

    cList<cDvbSink<Output> > sinks;
};

template<class Output>
inline cDvbSource<Output>::~cDvbSource()
{
    while (sinks.Count() > 0)
        sinks.Del(sinks.First(), false);
}

template<class Output>
inline void cDvbSource<Output>::EmitPackets(uchar const* Data, int Length)
{
    for (cDvbSink<Output>* Sink = sinks.First(); Sink != NULL; Sink = sinks.Next(Sink))
        Sink->ReceivePackets(Data, Length);
}

// --- cReceiverSource -------------------------------------------------------

template<class Output>
class cExternalSource : public cDvbSource<Output>
    ///< This provides a data source that receives data externally, e.g. when 
    ///< a thread calls cReceiverSource::EmitPackets.
{
public:
    cExternalSource() {}

    int PutPackets(uchar const* Data, int Length);

private:
    // cDvbSource Implementation
    virtual tDvbPacketInfos& GetPacketInfos() { return infos; }

private:
    tDvbPacketInfos infos;
};

template<class Output>
inline int cExternalSource<Output>::PutPackets(uchar const* Data, int Length)
{
    int Used;
    uchar const* p = GetTSPacketRange(Data, Length, Used);
    cDvbSource<Output>::EmitPackets(p, Length);
    infos.RangeOffset += Length;
    return Used;
}
#if 0
// --- cFileReader -----------------------------------------------------------

template<class Output>
class cFileReader : public cDvbSource<Output>
{
public:
    cFileReader(char const* FileName);

private:
    static void EmitFunc(uchar const* Data, int Length, void* Param);

    using cDvbSource<Output>::EmitPackets;

    cAsyncTSBuffer buffer;
    int fd;
};

template<class Output>
cFileSource<Output>::cFileSource(char const* FileName)
    : cThread("FileSource")
    , buffer(&cFileSource::EmitFunc, this, true)
    , fd(open(FileName, O_RDONLY))
{
    if (fd == -1)
        esyslog("ERROR: Couldn't open %s: %s", FileName, strerror(errno)); // XXX
}

template<class Output>
void cFileSource<Output>::EmitFunc(uchar const* Data, int Length, void* Param)
{
    cFileSource* source = static_cast<cFileSource*>(Param);
    source->EmitPackets(Data, Length);
}

template<class Output>
void cFileSource<Output>::Action()
{
    if (fd == -1) return;

    uchar data[BUFSIZ];
    int length = 0;
    while (Running() && (length = read(fd, data, BUFSIZ)) > 0)
        buffer.Put(data, length);
    if (length < 0)
        esyslog("ERROR: Couldn't read file: %s", strerror(errno)); // XXX
}
#endif
#endif // __DVBSOURCE_H
